@@ -21,6 +21,13 @@ import popen2 # TODO: use subprocess
 from xml.dom import minidom
 
 
+VERSION = '1.8.0'
+
+
+#
+# Helpers
+#
+
 def revs(rev):
     """
     Make sense out of convenient way of mentioning revisions
@@ -218,6 +225,66 @@ def determinebranch(branch, path, svninfo=svninfo):
     return '/'.join(new_chunks)
 
 
+def determinetag(tagname, path, svninfo=svninfo):
+    """
+    Let's set up a dummy 'svn info' command handler:
+
+      >>> def dummyinfo(path):
+      ...     return '''\
+      ... Path: .
+      ... URL: http://dev.worldcookery.com/svn/bla/trunk/blergh
+      ... Repository UUID: ab69c8a2-bfcb-0310-9bff-acb20127a769
+      ... Revision: 1654
+      ... Node Kind: directory
+      ... '''
+
+    ``determinetag()`` takes the svn path of the current working
+    directory path and mangles it to point to a named tag.
+    Here's an example:
+
+      >>> determinetag('foobar', '.', svninfo=dummyinfo)
+      'http://dev.worldcookery.com/svn/bla/tags/foobar/blergh'
+
+    We can do the same with branches
+
+      >>> def dummyinfo(path):
+      ...     return '''\
+      ... Path: .
+      ... URL: http://dev.worldcookery.com/svn/bla/branches/foobar/blergh
+      ... Repository UUID: ab69c8a2-bfcb-0310-9bff-acb20127a769
+      ... Revision: 1654
+      ... Node Kind: directory
+      ... '''
+
+      >>> determinetag('foobaz', '.', svninfo=dummyinfo)
+      'http://dev.worldcookery.com/svn/bla/tags/foobaz/blergh'
+ 
+    """
+    lines = svninfo(path).splitlines()
+    if lines[1].startswith('URL: '):
+        url = lines[1][5:]
+    else:
+        url = lines[2][5:]
+
+    chunks = url.split('/')
+    chunks.reverse()
+    new_chunks = []
+
+    while chunks:
+        ch = chunks.pop()
+        if ch in ('branch', 'branches'):
+            chunks.pop()
+            new_chunks.append('tags')
+            new_chunks.append(tagname)
+        elif ch == 'trunk':
+            new_chunks.append('tags')
+            new_chunks.append(tagname)
+        else:
+            new_chunks.append(ch)
+
+    return '/'.join(new_chunks)
+
+
 def listbranches(path, svninfo=svninfo, svnls=svnls):
     r"""
     Let's set up a dummy 'svn info' command handler:
@@ -315,6 +382,31 @@ def branchpoints(branch, svnlog=svnlog):
             int(newest_entry.getAttribute('revision')))
 
 
+#
+# Command registration
+#
+
+
+COMMANDS = {}
+ALIASES = {}
+
+def command(cmd, help_msg, alias=None):
+    """Register a command."""
+    def decorator(fn):
+        COMMANDS[cmd] = fn
+        if alias:
+            ALIASES[alias] = fn
+        fn.help_msg = help_msg
+        fn.alias = alias
+        return fn
+    return decorator
+
+
+#
+# Commands
+#
+
+@command('merge', 'merge branches', alias="ezmerge")
 def ezmerge(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -394,6 +486,7 @@ def ezmerge(argv, progname=None):
         os.system(merge_cmd)
 
 
+@command('revert', 'revert checkins', alias="ezrevert")
 def ezrevert(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -436,6 +529,7 @@ def ezrevert(argv, progname=None):
         os.system(merge_cmd)
 
 
+@command('switch', 'switch to a different branch', alias="ezswitch")
 def ezswitch(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -497,6 +591,42 @@ def ezswitch(argv, progname=None):
         os.system(cmd)
 
 
+@command('tag', 'make tags')
+def eztag(argv, progname=None):
+    progname = progname or os.path.basename(argv[0])
+    parser = optparse.OptionParser(
+                "usage: %prog [-n] [-m MSG] newtagname",
+                prog=progname,
+                description="Make a Subversion tag.")
+    parser.add_option('-m',
+                      help='commit message',
+                      action='store', dest='message', default=None)
+    parser.add_option('-n', '--dry-run',
+                      help='do not make the tag, just print the command',
+                      action='store_true', dest='dry_run', default=False)
+    try:
+        opts, args = parser.parse_args(argv[1:])
+        if len(args) > 2:
+            parser.error("too many arguments, try %s --help" % progname)
+    except optparse.OptParseError, e:
+        sys.exit(e)
+
+    # TODO: allow a different wc-path
+    path = '.'
+
+    if len(args) < 1:
+        parser.error("too few arguments, try %s --help" % progname)
+    newtag = determinetag(args[0], path)
+
+    cmd = "svn cp %s %s" % (path, newtag)
+    if opts.message:
+        cmd += " -m '%s'" % opts.message
+    print cmd
+    if not opts.dry_run:
+        os.system(cmd)
+
+
+@command('branchurl', 'print full URL of a branch', alias="ezbranch")
 def ezbranch(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -533,6 +663,7 @@ def ezbranch(argv, progname=None):
     print determinebranch(branch, path)
 
 
+@command('rmbranch', 'remove branches')
 def rmbranch(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -575,6 +706,7 @@ def rmbranch(argv, progname=None):
         os.system(cmd)
 
 
+@command('mvbranch', 'rename branches')
 def mvbranch(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -618,6 +750,7 @@ def mvbranch(argv, progname=None):
         os.system(cmd)
 
 
+@command('branchdiff', 'show combined diff of all changes on a branch')
 def branchdiff(argv, progname=None):
     progname = progname or os.path.basename(argv[0])
     parser = optparse.OptionParser(
@@ -646,6 +779,7 @@ def branchdiff(argv, progname=None):
     os.system(diff_cmd)
 
 
+@command('selftest', 'run self-tests')
 def selftest(argv, progname=None):
     import doctest
     failures, tests = doctest.testmod()
@@ -653,40 +787,31 @@ def selftest(argv, progname=None):
         print "All %d tests passed." % tests
 
 
+@command('help', 'this help message')
 def help(argv, progname=None):
     progname = os.path.basename(argv[0])
     print "usage: %s command arguments" % progname
     print "where command is one of"
-    print "  switch     -- switch to a different branch"
-    print "  merge      -- merge branches"
-    print "  revert     -- revert checkins"
-    print "  rmbranch   -- remove branches"
-    print "  branchurl  -- print full URL of a branch"
-    print "  branchdiff -- show combined diff of all changes on a branch"
-    print "  selftest   -- run self-tests"
-    print "  help       -- this help message"
+    width = max(map(len, COMMANDS))
+    for cmd, fn in sorted(COMMANDS.items()):
+        if fn.alias:
+            alias = ' (aka %s)' % fn.alias
+        else:
+            alias = ''
+        print "  %s -- %s%s" % (cmd.ljust(width), fn.help_msg, alias)
     print "Use %s command --help for more information about commands" % progname
 
 
+#
+# Dispatch
+#
+
 def eazysvn(argv):
     progname = os.path.basename(argv[0])
-    commands = {
-        'merge': ezmerge,
-        'revert': ezrevert,
-        'switch': ezswitch,
-        'rmbranch': rmbranch,
-        'mvbranch': mvbranch,
-        'branchurl': ezbranch,
-        'branchdiff': branchdiff,
-        'selftest': selftest,
-        'help': help,
-        '-h': help,
-        '--help': help,
-        }
-    if len(argv) < 2:
+    if len(argv) < 2 or argv[1] in ('-h', '--help'):
         return help(argv)
     cmd = argv.pop(1)
-    func = commands.get(cmd)
+    func = COMMANDS.get(cmd)
     if func is None:
         sys.exit("Unknown command: %s." % cmd)
     progname = progname + ' ' + cmd
@@ -694,17 +819,20 @@ def eazysvn(argv):
 
 
 def main():
+    if len(sys.argv) > 1 and sys.argv[1] == '--version':
+        print "eazysvn version %s" % VERSION
+        sys.exit(0)
     cmd = os.path.basename(sys.argv[0])
     if cmd.endswith('.py'):
         cmd = cmd[:-len('.py')]
-    commands = {
-        'ezmerge': ezmerge,
-        'ezswitch': ezswitch,
-        'ezbranch': ezbranch,
-        'ezrevert': ezrevert,
-        }
-    func = commands.get(cmd, eazysvn)
+    func = ALIASES.get(cmd, eazysvn)
     sys.exit(func(sys.argv))
 
+
+def additional_tests(): # for setup.py test
+    import doctest
+    return doctest.DocTestSuite()
+
+
 if __name__ == '__main__':
     main()
